from datetime import datetime, timedelta


class Prayer:
    def __init__(self, name: str, time: datetime) -> None:
        self.name = name
        self.time = time

    def has_passed(self) -> bool:
        return datetime.now() > self.time

    def time_left(self) -> timedelta:
        return self.time - datetime.now()

    def __str__(self) -> str:
        return f"{self.name}: {self.time.strftime('%H:%M')}"



from typing import Optional
from urllib.parse import quote


class City:
    def __init__(self, city: str, country: str, state: Optional[str] = None) -> None:
        self.city = city
        self.state = state
        self.country = country


class Coordinates:
    def __init__(self, latitude: float, longitude: float) -> None:
        self.latitude = latitude
        self.longitude = longitude


class Address:
    def __init__(self, address: str) -> None:
        self.address = address

    def url_encoded(self) -> str:
        return quote(self.address, safe="")



# Purpose: Contains the Method class, which is used to represent a method of
# calculating prayer times.


from myprayer.api.location_types import Coordinates


class Method:
    name: str
    id: int
    params: dict[str, str]
    location: Coordinates

    def __init__(
        self,
        data: dict,
    ):
        self.name = data["name"]
        self.id = int(data["id"])
        self.params = data["params"]
        if "location" in data:
            self.location = Coordinates(
                data["location"]["latitude"],
                data["location"]["longitude"],
            )

    def __str__(self) -> str:
        return f"{self.id}: {self.name}"



from datetime import datetime, timedelta

from myprayer.api.prayer import Prayer


class Day:
    """Represents a single day with prayer times.

    Attributes:
        day (int): The day number in the month
        month (int): The month number 1-12
        year (int): The year
        data (dict): The prayer time data for this day
        prayers (list[Prayer]): List of Prayer objects
        skip (list[str]): Prayer names to skip

    Methods:
        get_next_prayer(): Returns the next prayer that has not passed yet
        get_prayer(name): Returns the Prayer object with the given name
        has_passed(): Checks if the last prayer of the day has passed

    Raises:
        ValueError: If day is not 1-31 or month is not 1-12

    Examples:
        >>> data = Client().get_day(15, 1, 2022)
        >>> print(day.prayers[0].name)
        Fajr

        >>> day.get_prayer("fajr").__str__()
        Fajr: 05:20
    """

    day: int
    month: int
    year: int
    data: dict
    prayers: list[Prayer]

    def __init__(
        self, day: int, month: int, year: int, data: dict, skip: list[str] = []
    ):
        if day not in range(1, 32):
            raise ValueError("Day must be in range 1 to 31")
        if month not in range(1, 13):
            raise ValueError("Month must be in range 1 to 12")

        self.data: dict = data
        self.day: int = day
        self.month: int = month
        self.year: int = year
        self.date: datetime = datetime(year, month, day)
        self.prayers: list[Prayer] = []
        self.skip: list[str] = skip

        timings = data["timings"]
        day_passed = False
        last_prayer = None
        for prayer, time in timings.items():
            if prayer in self.skip:
                continue
            prayer_time = datetime.strptime(time[:5], "%H:%M").replace(
                day=self.day,
                month=self.month,
                year=self.year,
            )
            if last_prayer and (prayer_time - last_prayer).total_seconds() < 0:
                day_passed = True

            if day_passed:
                prayer_time += timedelta(days=1)
            self.prayers += [Prayer(prayer, prayer_time)]
            last_prayer = prayer_time

    def get_next_prayer(self) -> Prayer | None:
        for prayer in self.prayers:
            if not prayer.has_passed():
                return prayer

    def get_prayer(self, name: str) -> Prayer | None:
        for prayer in self.prayers:
            if prayer.name == name:
                return prayer

    def has_passed(self) -> bool:
        if self.prayers[-1].has_passed():
            return True
        return False



# Description: This file contains all the constants used in the project

import os
from pathlib import Path
from typing import Dict, Final, List

from myprayer.cli.enums import NextOutType, OutType, TimeFormat

APP_NAME: Final[str] = "myprayer"
# Prayer times API URL

LOCATION_TYPES: Final[List[str]] = ["City", "Coordinates", "Address"]

# Find config/cache dir based on OS
if str(os.name) == "nt":
    cache_dir = (
        Path(os.environ.get("LOCALAPPDATA") or Path.home() / "AppData/Local") / APP_NAME
    )
    config_dir = (
        Path(os.environ.get("APPDATA") or Path.home() / "AppData/Roaming") / APP_NAME
    )
else:
    cache_dir = (
        Path(os.environ.get("XDG_CACHE_HOME") or Path.home() / ".cache") / APP_NAME
    )
    config_dir = (
        Path(os.environ.get("XDG_CONFIG_HOME") or Path.home() / ".config") / APP_NAME
    )

# cache dir path
CACHE_DIR: Final[Path] = cache_dir

# Create cache dir if it doesn't exist
CACHE_DIR.mkdir(parents=True, exist_ok=True)

# config dir path
CONFIG_DIR: Final[Path] = config_dir

# config file path
CONFIG_FILE: Final[Path] = CONFIG_DIR / "config.json"


# File format for cache files
FILE_FORMAT: Final[str] = "{country}_{city}_{month}_{year}_{method}.json"

# Create list for prayer names
PRAYERS: Final[List[str]] = [
    "Fajr",
    "Sunrise",
    "Dhuhr",
    "Asr",
    "Maghrib",
    "Isha",
    "Midnight",
]

DEFAULT_PRAYERS: Final[List[str]] = ["Fajr", "Dhuhr", "Asr", "Maghrib", "Isha"]

# Create dict for calculation methods
CALCULATION_METHODS: Final[Dict[str, int]] = {
    "University of Islamic Sciences, Karachi": 1,
    "Islamic Society of North America": 2,
    "Muslim World League": 3,
    "Umm Al-Qura University, Makkah": 4,
    "Egyptian General Authority of Survey": 5,
    "Institute of Geophysics, University of Tehran": 7,
    "Gulf Region": 8,
    "Kuwait": 9,
    "Qatar": 10,
    "Majlis Ugama Islam Singapura, Singapore": 11,
    "Union Organization islamic de France": 12,
    "Diyanet İşleri Başkanlığı, Turkey": 13,
    "Spiritual Administration of Muslims of Russia": 14,
}

# dict for time formats (strftime/strptime)
TIME_FORMATS: Final[Dict[TimeFormat, str]] = {
    TimeFormat.twelve: "%I:%M %p",
    TimeFormat.twenty_four: "%H:%M",
}

# dict for timedelta
TIMEDELTA: Final[Dict[OutType | NextOutType, str]] = {
    OutType.pretty: "{hours}H {minutes}M",
    OutType.machine: "{hours:02d}H{minutes:02d}M",
    OutType.table: "{hours}H {minutes}M",
    OutType.json: "{hours}H {minutes}M",
    NextOutType.waybar: "{hours}H {minutes}M",
}


# Waybar icons
WAYBAR_ICONS = {
    "Fajr": "󰖜",
    "Dhuhr": "󰖙",
    "Asr": "󰼰",
    "Maghrib": "󰖛",
    "Isha": "󰖔",
}



from calendar import month_name, monthrange

from myprayer.api.day import Day


class Month:
    """A class representing a calendar month.

    Attributes:
        data (dict): The prayer time data for each day of the month
        name (str): The name of the month (e.g. "January"), derived from the month number
        month (int): The number of the month, 1-12
        year (int): The year
        days (int): The number of days in the month, derived from the month and year
        skip (list[str]): Prayers to skip when getting prayer times

    Methods:
        get_day(day): Returns a Day object for the given day number
        has_passed(): Checks if the last day of the month has passed, i.e. if the last prayer time in the last day has passed

    Raises:
        ValueError: If month is not 1-12 or day is invalid for the month

    Examples:
        >>> jan = Client().get_month(1, 2022)
        >>> jan.name
        "January"

        >>> day = jan.get_day(15)
        >>> day.get_prayer("fajr").__str__()
        Fajr: 05:20

        >>> jan.has_passed()
        False
    """

    data: dict
    name: str
    month: int
    year: int
    days: int

    def __init__(
        self,
        month: int,
        year: int,
        data: dict,
        skip: list[str] = [],
    ):
        if month not in range(1, 13):
            raise ValueError("Month must be in range 1 to 12")

        self.name = month_name[month]
        self.month = month
        self.year = year
        self.data = data
        self.skip = skip
        self.days = monthrange(year, month)[1]

    def get_day(self, day: int) -> Day:
        if day not in range(1, self.days + 1):
            raise ValueError(f"{self.name} {self.year} is only {self.days} days long.")

        return Day(day, self.month, self.year, self.data[day - 1], self.skip)

    def has_passed(self) -> bool:
        # Check if the last day of the month has passed.
        return self.get_day(self.days).has_passed()



import json
from datetime import datetime
from pathlib import Path
from typing import Optional

import requests

from myprayer.api.day import Day
from myprayer.api.exceptions import *
from myprayer.api.location_types import Address, City, Coordinates
from myprayer.api.method import Method
from myprayer.api.month import Month
from myprayer.api.prayer import Prayer


API = "http://api.aladhan.com/v1/"
SKIP = ["Firstthird", "Lastthird", "Imsak", "Sunset"]


class Client:
    """
    Client for accessing the Aladhan prayer times API.

    Attributes:
        location: The location to get prayer times for. Can be a City,
            Coordinates, or Address object.
        method: The calculation method to use. Optional, defaults to auto.
        skip: Prayers to skip from the results. e.g. ["Sunrise", "Sunset"]. Optional.
        cache_dir: Optional directory to cache API responses.

    Methods:
        get_today: Get prayer times for today.
        get_next_prayers: Get prayer times for the next prayer after the
            current time.
        get_methods: Get list of available calculation methods.
        get_day: Get prayer times for a specific day.
        get_month: Get prayer times for a whole month.

    The Client handles querying the API, caching responses, and raising
    any API errors.
    """

    location: City | Coordinates | Address
    method: Optional[int]
    skip: list[str]
    cache_dir: Optional[Path]

    def __init__(
        self,
        location: City | Coordinates | Address,
        method: Optional[int | Method] = None,
        skip: list[str] = [],
        cache_dir: Optional[Path | str] = None,
    ) -> None:
        self.location = location
        if isinstance(method, Method):
            self.method = method.id
        else:
            self.method = method
        self.skip = skip + SKIP

        if cache_dir is not None:
            self.cache_dir = (
                Path(cache_dir) if isinstance(cache_dir, str) else cache_dir
            )
            self.cache_dir.mkdir(parents=True, exist_ok=True)

    @staticmethod
    def __set_location_params(params: dict, location: City | Coordinates | Address):
        if isinstance(location, Coordinates):
            params["latitude"] = location.latitude
            params["longitude"] = location.longitude
        elif isinstance(location, Address):
            params["address"] = location.url_encoded()
        elif isinstance(location, City):
            params["city"] = location.city
            params["country"] = location.country
            if location.state:
                params["state"] = location.state

    @staticmethod
    def __get_endpoint_by_location(
        location: City | Coordinates | Address, endpoint: str
    ):
        if isinstance(location, Coordinates):
            return endpoint
        if isinstance(location, Address):
            return endpoint + "ByAddress"
        if isinstance(location, City):
            return endpoint + "ByCity"

    @staticmethod
    def __detect_exceptions(response: requests.Response, data: dict):
        if response.status_code == 429:
            raise RateLimitedException("Rate limit exceeded.")
        if response.status_code == 400:
            raise BadRequestException(data["data"])
        if response.status_code == 500:
            raise ServerErrorException(data["message"])

    # TODO: add support for different methods
    @staticmethod
    def __get_cache_file_name(
        month: int,
        year: int,
        location: City | Coordinates | Address,
        method: Optional[int] = None,
    ):
        if method is None:
            method = "auto"  # type: ignore
        suffix = f"{month}_{year}_{method}.json"
        if isinstance(location, Coordinates):
            return f"coordinates_{location.longitude}_{location.latitude}_{suffix}"
        if isinstance(location, Address):
            return f"address_{location.url_encoded()}_{suffix}"
        if isinstance(location, City):
            return f"city_{location.city.replace(' ','')}_{location.country.replace(' ', '')}_{suffix}"

    @staticmethod
    def get_methods() -> list[Method]:
        ENDPOINT = "methods"
        response = requests.get(API + ENDPOINT)
        data = response.json()
        Client.__detect_exceptions(response, data)
        methods = [
            Method(method) for method in data["data"].values() if method["id"] != 99
        ]
        methods.sort(key=lambda method: method.id)
        return methods

    def get_today(self) -> Day:
        today = datetime.today()
        return self.get_day(today.day, today.month, today.year)

    def get_next_prayers(self) -> Day:
        # Get the current month and day.
        today = datetime.today()
        current_month = self.get_month(today.month, today.year)
        current_day = current_month.get_day(today.day)

        # Check if the current day has already passed.
        if current_day.has_passed():
            # Check if the current month has passed.
            if current_month.has_passed():
                # If it's December, switch to the next year, and set the month to January.
                if current_month.month == 12:
                    next_month = self.get_month(1, today.year + 1)
                else:
                    # Otherwise, switch to the next month in the current year.
                    next_month = self.get_month(today.month + 1, today.year)
                next_day = next_month.get_day(1)
            else:
                # If the month hasn't passed, just move to the next day in the same month.
                next_day = current_month.get_day(today.day + 1)
        else:
            # If the current day hasn't passed, keep it as is.
            next_day = current_day

        return next_day

    def get_next_prayer(self) -> Prayer | None:
        return self.get_next_prayers().get_next_prayer()

    def get_day(self, day: int, month: int, year: int) -> Day:
        return self.get_month(month, year).get_day(day)

    def get_month(self, month: int, year: int) -> Month:
        if self.cache_dir:
            try:
                data = self.load(month, year)
            except FileNotFoundError:
                data = self.fetch(month, year)
                self.save(data, month, year)
        else:
            data = self.fetch(month, year)

        return Month(month, year, data, self.skip)

    def fetch(self, month: int, year: int) -> dict:
        # define parameters
        params = {
            "month": month,
            "year": year,
        }
        # check if a method was provided
        if self.method is not None:
            params["method"] = self.method

        if self.location is None:
            raise ValueError("No location was provided.")
        # set the location type and the ENDPOINT
        self.__set_location_params(params, self.location)
        ENDPOINT = self.__get_endpoint_by_location(self.location, "calendar")

        # send the request and get the JSON data
        response = requests.get(API + ENDPOINT, params=params)
        data = response.json()
        self.__detect_exceptions(response, data)

        return data["data"]

    def save(self, data: dict, month: int, year: int) -> None:
        if self.cache_dir is None:
            raise ValueError("No cache directory was provided.")
        self.cache_dir.mkdir(parents=True, exist_ok=True)
        with open(
            self.cache_dir
            / self.__get_cache_file_name(month, year, self.location, self.method),
            "w",
        ) as f:
            json.dump(data, f)

    def load(self, month, year) -> dict:
        if self.cache_dir is None:
            raise ValueError("No cache directory was provided.")
        if self.cache_dir.exists() is False:
            raise FileNotFoundError("Cache directory does not exist.")
        with open(
            self.cache_dir
            / self.__get_cache_file_name(month, year, self.location, self.method),
            "r",
        ) as f:
            return json.load(f)


if __name__ == "__main__":
    client = Client(
        City("London", "United Kingdom"),
        cache_dir=Path("cache"),
    )
    for prayer in client.get_next_prayers().prayers:
        print(prayer)
    print()
    method = Client.get_methods()[5]
    client = Client(
        City("Cairo", "Egypt"),
        skip=["Sunrise", "Sunset"],
        cache_dir=Path("cache"),
        method=method,
    )
    for prayer in client.get_next_prayers().prayers:
        print(prayer)

    jan = client.get_month(1, 2022)
    day = jan.get_day(15)
    print(day.get_prayer("Fajr").__str__())
